#include <kdebug.h>

#include "syncml.h"
#include "syncmlresponsepacket.h"
#include "mlhttpserver.h"

#include "syncmldevinf.h"
#include "syncmldatabase.h"
#include "mlsessionhandler.h"

#include "xmlbeautifier.h"


SMLSessionHandler::SMLSessionHandler(MLHttpServer * communication_, SMLVersion protocolVersion) : 
	URI(),remoteURI(),pending(),databaseMap(), syncmlVersion(protocolVersion), communication(communication_)
{
	response = new SMLResponsePacket();
	DevInf=NULL;
	remoteDevInf=NULL;
	CmdNoDevInf.MsgRef = 0xf000; // Failry safe value to use since a
		// usuall session consists of 7
		// packets total
	CmdNoDevInf.CmdRef=0;
	pending[CmdNoDevInf] = SMLError::NoDevInf;
	
	DevInf = new SMLDevInf();
	
	automaticResponse = true;
	everythingFinishedFlag = false;
};

SMLSessionHandler::~SMLSessionHandler() {
  delete DevInf;
  delete remoteDevInf;
  delete response;
}

SMLDevInf* SMLSessionHandler::getDevInf() { 
  return DevInf; 
}

void SMLSessionHandler::forceSend()
{
	automaticResponse = true;
	getResponse()->forceSend();
}

void SMLSessionHandler::slotParsingFinished()
{
	kdDebug() << "SMLSessionHandler: Responding to the client?" << endl;
	if (automaticResponse)
	{
		kdDebug() << "SMLSessionHandler: Yes, responding to the client." << endl;

		QByteArray data;
		
		QTextStream * outputStream = new QTextStream(data, IO_WriteOnly);
		SMLFlattener *s = new SyncMLFlattener(outputStream, getResponse()->getNamespaceContainer());
		getResponse()->sourceDevice()->setURI(getURI());
		getResponse()->targetDevice()->setURI(getRemoteURI());
		//handler->sendAlerts();
		getResponse()->send(s);
		outputStream->device()->flush();
	
		delete s;
		delete outputStream;

		communication->sendData(data);
		automaticResponse = false;

		if (everythingFinishedFlag)
		{
			//XXX FIXME: delete session here.
			kdDebug() << "SMLSessionHandler: Disconnecting the client." << endl;
			QObject::connect(communication, SIGNAL(clientDisconnected()),
					 this, SLOT(slotEverythingDisconnected()));

			communication->disconnectClient();
		}
	}
	else
	{
		kdDebug() << "SMLSessionHandler: No, not yet." << endl;
	}
}

void SMLSessionHandler::slotEverythingFinished()
{
	//we need to wait until the last packet is sent
	everythingFinishedFlag = true;
}

void SMLSessionHandler::slotEverythingDisconnected()
{
	kdDebug() << "SMLSessionHandler: Client disconnected. Bye bye." << endl;
	emit sessionFinished();	
}

void SMLSessionHandler::readDatabases()
{
	//it's empty for now.
}

void SMLSessionHandler::addDatabase(KSyncML::SyncMLDatabase *database) 
{
	QString qDatabaseName = database->getURI().getQAbsoluteLocURI();
	databaseMap[qDatabaseName] = database;
	kdDebug() << "SMLSessionHandler(" << (void *)this << "): Adding database " 
			<< (void *)database << " with uri : " << database->getURI().getLocURI().c_str() 
			<< " (absoluteURI=" << qDatabaseName << ")" << endl;
		

	//XXX FIXME we need to copy the datastore info to the main devinf
	//in order to let it advertise. this should be regulated by some preferences settings
	KSyncML::SyncMLDatabase::DataStoreType * store = database->getDataStoreInfo();
	store->SourceRef = SMLURI(store->SourceRef,"", getURI()).getLocURI();
	getDevInf()->addDataStoreInfo(*store);
}

bool SMLSessionHandler::hasDatabase(const SMLURI & uri)
{
	QString name = uri.getQLocURI(false);
	return (databaseMap.contains(name));
}


void SMLSessionHandler::slotDataReceived(SyncMLSession::TSynceesList& data, SMLURI& databaseURI)
{
	QString databaseName = databaseURI.getQAbsoluteLocURI();
	kdDebug() << "SMLSessionHandler(" << (void *)this << "): received data for database: " << databaseName << endl;
	
	if (databaseMap.contains(databaseName)) 
	{
		KSyncML::SyncMLDatabase * database = databaseMap[databaseName];
		
		kdDebug() << "SMLSessionHandler(" << (void *)this << "): found database: " << (void *)database << endl;
		database->processClientData(data);
		
		//
		// XXX no. this should wait the end of parsing of xml before firing the signal.
		//
		emit databaseDataRead(database);
		return;
	}
	else
	{
		kdWarning() << "SMLSessionHandler: received data for an unregistered database: " << databaseName << endl;
	}
}

void SMLSessionHandler::slotKeysReceived(SyncMLSession::TKeysList& data, SMLURI& databaseURI)
{
	QString databaseName = databaseURI.getQAbsoluteLocURI();
	kdDebug() << "SMLSessionHandler: received keys for database: " << databaseName << endl;
	
	KSyncML::SyncMLDatabase * database = databaseMap[databaseName];
	if (database) 
	{
		database->processClientKeys(data);
	}
	else
	{
		kdWarning() << "SMLSessionHandler: received keys for an unregistered database: " << databaseName << endl;
	}
}


void SMLSessionHandler::writeDatabases()
{
	//XXX dammit, there is a nil pointer with complete URI in the map. 
	//Of course! Check the QMap doc for explanation about [] operator
	int databasesWritten = 0;
	kdDebug() << "SMLSessionHandler(" << (void *)this << "): writing the databases." << endl;
	KSyncML::SyncMLDatabaseMapType::Iterator it;
	
	for ( it = databaseMap.begin(); it != databaseMap.end(); ++it ) 
	{
		QString databaseName = it.key();
		kdDebug() << "SMLSessionHandler: write data for database: " << databaseName << endl;
		KSyncML::SyncMLDatabase * database = it.data();
		//kdDebug() << "SMLSessionHandler: writing data for database: " << (void *)database << endl;
		if (database)
		{
			//database->writeData();
			
			sendDatabaseChangeLog(database);
			
			kdDebug() << "SMLSessionHandler: written data for database: " << databaseName << endl;
			emit databaseDataWritten(database);
			databasesWritten++;
		}
	}

	automaticResponse = true;	
	if (databasesWritten)
	{
		slotParsingFinished();
	}
}


void SMLSessionHandler::sendDatabaseChangeLog(KSyncML::SyncMLDatabase * database)
{
	KSyncML::ChangeLog *cLog = database->getDBChangeLog();
	
	if (cLog == NULL)
	{
		return;
	}
	
	QValueList<KSyncML::ChangeLog::Record *>::iterator it;
	QValueList<KSyncML::ChangeLog::Record *> records = cLog->recordsList();

	//
	// At first, we send all the "OK"s
		
	for ( it = records.begin(); it != records.end(); ++it )
	{
		KSyncML::ChangeLog::Record * r = *it;
		SyncMLItem * item = r->item;

		if ((r->action == KSyncML::ChangeLog::OK) || (r->action == KSyncML::ChangeLog::REPLACE))
		{
			//XXX FIXME: Replace "Add" and "Replace" with a token translated via WBX
			QString cmd = "Add";
			if (item->action == SyncMLItem::REPLACED)
				cmd = "Replace";
			
			//SMLTreeNode* t = response->addStatus(item->msgRef, item->cmdRef, cmd, SMLError::ItemAdded);
			SMLTreeNode* t = response->addStatus(item->msgRef, item->cmdRef, cmd, SMLError::Successful);
			new SMLTreeNode(t,SMLToken::SourceRef, item->source);
		}
	}
	
	
	//
	// Then, we add a "Sync" command
	SMLCmdRef syncRef;
	//We must have a "sync" node for the reply.
	SMLTreeNode* syncNode = response->addCommand(SMLToken::Sync, &syncRef);
	database->getURI().addToNode(new SMLTreeNode(syncNode,SMLToken::Source));
	//SMLURI target(item->target.latin1());
	
	//XXX FIXME: WTF IS THAT??
	SMLURI target(QString("./c:\\system\\data\\Contacts.cdb"));//P900 debugging
	target.addToNode(new SMLTreeNode(syncNode,SMLToken::Target));

	
	//
	// And we add all the missing records
	for ( it = records.begin(); it != records.end(); ++it )
	{
		KSyncML::ChangeLog::Record * r = *it;
		SyncMLItem * item = r->item;

		if ((r->action == KSyncML::ChangeLog::ADD) || (r->action == KSyncML::ChangeLog::REPLACE))
		{
			SMLCmdRef ret;
			SMLTreeNode *addNode = NULL;
			if (r->action == KSyncML::ChangeLog::REPLACE)
				addNode = response->addSubCommand(syncNode, SMLToken::Replace, &ret, true); 
			else
				addNode = response->addSubCommand(syncNode, SMLToken::Add, &ret, true); 
				
			SMLType vCard_("text/x-vcard");
			vCard_.addToNode(addNode);
			
			SMLTreeNode *itemNode = new SMLTreeNode(addNode, SMLToken::Item);
			
			if (item->target != "")
			{
				string tempSTLStr(item->target.latin1());
				SMLURI temp(tempSTLStr);
				temp.addToNode(new SMLTreeNode(itemNode,SMLToken::Target));
			}
			else
			{
				string tempSTLStr(item->source.latin1());
				SMLURI temp(tempSTLStr);
				temp.addToNode(new SMLTreeNode(itemNode,SMLToken::Source));
			}
			new SMLTreeNode(itemNode,SMLToken::Data, item->item);
		}
	}
	delete cLog;
}


SMLCmdRef SMLSessionHandler::sendTwoWaySync(SMLURI source, SMLURI target, string next) {
  //XXX FIXME why addBase is called?
  //source.addBase(getURI());
  //target.addBase(getRemoteURI());
  SMLCmdRef  ret;
  SMLTreeNode *t = response->addCommand(SMLToken::Alert,&ret);
  new SMLTreeNode(t,SMLToken::Data,QString("200"));
  SMLTreeNode *i = new SMLTreeNode(t,SMLToken::Item);
  target.addToNode(new SMLTreeNode(i,SMLToken::Target));
  source.addToNode(new SMLTreeNode(i,SMLToken::Source));
  SMLTreeNode *a = new SMLTreeNode(new SMLTreeNode(i,SMLToken::Meta),SMLToken::MetInf::Anchor);

  string last="";
//  if (databaseMap[source.getQAbsoluteLocURI()]) 
 //   last = databaseMap[source.getQAbsoluteLocURI()]->lastSyncWith(this,target).ascii();

  if (last=="") last="00000000T000000Z";
  //                  12345678 123456

  new SMLTreeNode(a,SMLToken::MetInf::Last,QString(last.c_str()));
  if (next=="") 
  	next = string(SMLTimestamp().ascii());	
  new SMLTreeNode(a,SMLToken::MetInf::Next,QString(next.c_str()));
  pending[ret]=SMLError::Pending;
  return ret;
}

SMLCmdRef SMLSessionHandler::sendOneWaySync(SMLURI source, SMLURI target, string next) {
  //XXX FIXME why addBase is called?
  //source.addBase(getURI());
  //target.addBase(getRemoteURI());
  SMLCmdRef  ret;
  SMLTreeNode *t = response->addCommand(SMLToken::Alert,&ret);
  new SMLTreeNode(t,SMLToken::Data,QString("204"));
  SMLTreeNode *i = new SMLTreeNode(t,SMLToken::Item);
  
  target.addToNode(new SMLTreeNode(i,SMLToken::Target));
  source.addToNode(new SMLTreeNode(i,SMLToken::Source));
  SMLTreeNode *a = new SMLTreeNode(new SMLTreeNode(i,SMLToken::Meta),SMLToken::MetInf::Anchor);

  string last="";
//  if (databaseMap[source.getQAbsoluteLocURI()]) 
//    last = databaseMap[source.getQAbsoluteLocURI()]->lastSyncWith(this,target).ascii();

  if (last=="") last="00000000T000000Z";
  //                  12345678 123456

  new SMLTreeNode(a,SMLToken::MetInf::Last, QString(last.c_str()));
  if (next=="") 
  	next = string(SMLTimestamp().ascii());	
  new SMLTreeNode(a,SMLToken::MetInf::Next,QString(next.c_str()));
  pending[ret]=SMLError::Pending;
  return ret;
}


SMLCmdRef SMLSessionHandler::sendDevInf() 
{
	SMLCmdRef ret;
	if (DevInf!=NULL) 
	{
		SMLTreeNode *t = response->addCommand(SMLToken::Put,&ret);
		
		//XXX FIXME it's not a good place to change the protocol version...
		DevInf->setVerDTD(syncmlVersion);
		DevInf->addDevInfToPacket(t);
		pending[ret]=SMLError::Pending;
	} 
	else 
		ret = CmdNoDevInf;
	return ret;
}

SMLCmdRef SMLSessionHandler::requestDevInf() 
{
	SMLCmdRef ret;
	SMLTreeNode *t = response->addCommand(SMLToken::Get,&ret);
	new SMLTreeNode(new SMLTreeNode(t,SMLToken::Meta),SMLToken::MetInf::Type,SMLMimeType::devinfxml);

	QString verStr;  
	if (syncmlVersion == ONE_ZERO)
		verStr = "./devinf10";
	else
		verStr = "./devinf11";
	
	SMLURI d(verStr,"",&remoteURI);
	d.addToNode(new SMLTreeNode(new SMLTreeNode(t,SMLToken::Item),SMLToken::Target));
	
	pending[ret] = SMLError::Pending;
	return ret;
}

SMLErrorType SMLSessionHandler::getResult(SMLCmdRef ref) {
  return pending[ref];
}
/*
SMLCmdRef SMLSessionHandler::sendAdd(SMLTreeNode *syncNode,SMLType *type, SMLURI *source, SMLURI *target, string data) {
  SMLCmdRef ret;
  SMLTreeNode *t = response->addSubCommand(syncNode,SMLToken::Add,&ret);
  if (type) type->addToNode(t);
  SMLTreeNode *i = new SMLTreeNode(t,SMLToken::Item);
  if (source) source->addToNode(i);
  if (target) target->addToNode(i);
  new SMLTreeNode(i,SMLToken::Data,QString(data.c_str()));
  return ret;
}

SMLCmdRef SMLSessionHandler::sendCopy(SMLTreeNode *syncNode, SMLType *type, SMLURI *source, SMLURI *target) {
  // FIXME
  (void)syncNode;(void)type;(void)source;(void)target;
  SMLCmdRef a;
  return a;
}

SMLCmdRef SMLSessionHandler::sendDelete(SMLTreeNode *syncNode, SMLType *type, SMLURI *source, SMLURI *target) {
  // FIXME
  (void)syncNode;(void)type;(void)source;(void)target;  
  SMLCmdRef a;
  return a;
}

SMLCmdRef SMLSessionHandler::sendReplace(SMLTreeNode *syncNode, SMLType *type, SMLURI *source, SMLURI *target, string data) {
  // FIXME
  (void)syncNode;(void)type;(void)source;(void)target;(void)data;
  SMLCmdRef a;
  return a;
}

void SMLSessionHandler::sendAlerts()
{
	kdDebug() << "SMLSessionHandler: SEND ALERTS CALLED" << endl;
}
*/


void SMLSessionHandler::recDisplay(string what) 
{
	(void)what; 
}

void SMLSessionHandler::recDevInf() 
{
}

void SMLSessionHandler::setVer(SMLVersion newVer)
{
	if (syncmlVersion != newVer)
	{
		if (syncmlVersion != UNKNOWN)
			kdWarning() << "SMLSessionHandler: setting "
				<< "the protocol version more than once. You are looking for troubles." << endl;
		
		kdDebug() << "SMLSessionHandler: New version of protocol: " << newVer << endl;
	
		syncmlVersion = newVer;
		DevInf->setVerDTD(newVer);
		response->setSyncMLVersion(newVer);
	}
}


#include "mlsessionhandler.moc"


