/***************************************************************************
                                 libSyncML
           Implementation of the SyncML protocol (www.syncml.org)
         Copyright (C) 2001, 2002 by Max Berger <max.berger@xslt.de>
 **************************************************************************/

/***************************************************************************
 *                                                                         *
 *  This library is free software; you can redistribute it and/or          *
 *  modify it under the terms of the GNU Lesser General Public             *
 *  License as published by the Free Software Foundation; either           *
 *  version 2.1 of the License, or (at your option) any later              *
 *  version.                                                               *
 *                                                                         *
 *  This library is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU      *
 *  Lesser General Public License for more details.                        *
 *                                                                         *
 *  You should have received a copy of the GNU Lesser General Public       *
 *  License along with this library; if not, write to the Free             *
 *  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA      *
 *  02111-1307 USA                                                         *
 *                                                                         *
 ***************************************************************************/

#include <cstdlib>
#include <iomanip>
#include <vector>
#include <string>

#include <kdebug.h>

#include "syncml.h"
#include "syncmldevinf.h"
#include "xmlparser.h"
#include "xmlparser_tree.h"
#include "xmlnamespace.h"
#include "xmlbeautifier.h"



void SMLDevInf::realInit() {
  headNode = new SMLTreeNode(NULL,SMLToken::Root);
}

SMLDevInf::SMLDevInf() {
  realInit();
  enforceConstraints();
}

SMLDevInf::SMLDevInf(SMLTreeNode *t) {
  realInit();
  initFromTreeNode(t);
  enforceConstraints();
}

SMLDevInf::SMLDevInf(SMLDevInf *t) {
  realInit();
  initFromTreeNode(t->getDevInfTree());
  // enforceConstraints(); // not necessary since it's done in getdevinftree
}

void SMLDevInf::eventReceived(SMLTokenType event, SMLTreeNode *CurrentNode) {
  if (event == SMLToken::DevInf::DevInf) 
    initFromTreeNode(CurrentNode);
}

void SMLDevInf::initFromTreeNode(SMLTreeNode *otherRoot) {
  if (otherRoot->getType()==SMLToken::Root) 
    headNode->addChild( new SMLTreeNode(otherRoot->getChildren()->front()) );
  else
    headNode->addChild( new SMLTreeNode(otherRoot) );
}

void SMLDevInf::enforceConstraints() {
  enforceConstraints(headNode);
}

#define mustBeTextNode(node,default) \
  node->getChildren()->clear(); \
  if (node->getSTLValue()=="") node->setSTLValue(default)

  //XXX FIXME i don't understand the usefulness of such a function...
void SMLDevInf::enforceConstraints(SMLTreeNode *node) {

  // This is a must or we'd accidentally erase all attributes
  node->simplifyMe();
  // Now kill all textnodes. This is mainly for previous formatted text
  node->eliminateTextChildren();
  // DevInf doesn't specify any empty children. Therefore these values are
  // unset. to make sure, we kill any empty nodes
  // Note that mandatory notes will get re-created with default values
  node->eliminateEmptyChildren();

  if ((node->getType()) == SMLToken::Root) {
    SMLTokenType sub[] = {SMLToken::DevInf::DevInf};
    node->mustHaveChildren(1,sub);
  } else
  if ((node->getType()) == SMLToken::DevInf::DevId) {
    mustBeTextNode(node,"");
    if (node->getSTLValue()=="") setDevId();
  } else
  if ((node->getType()) == SMLToken::DevInf::DevInf) {
    
    // DataStore should be enforced! But this is one where we can't have any
    // reasonable defaults...
    SMLTokenType sub[] = {SMLToken::DevInf::VerDTD,
                          SMLToken::DevInf::DevId,
                          SMLToken::DevInf::DevTyp,
                          // SMLToken::DevInf::DataStore
    };
    node->mustHaveChildren(3,sub);

    SMLTokenType order[] = {SMLToken::DevInf::VerDTD,
                            SMLToken::DevInf::Man,
                            SMLToken::DevInf::Mod,
                            SMLToken::DevInf::OEM,
                            SMLToken::DevInf::FwV,
                            SMLToken::DevInf::SwV,
                            SMLToken::DevInf::HwV,
                            SMLToken::DevInf::DevId,
                            SMLToken::DevInf::DevTyp,
                            SMLToken::DevInf::DataStore,
                            SMLToken::DevInf::CTCap,
                            SMLToken::DevInf::Ext};
    node->orderChildren(12,order);
    
  } else
  if ((node->getType()) == SMLToken::DevInf::DevTyp) {
    mustBeTextNode(node,"workstation");
  } else
  if ((node->getType()) == SMLToken::DevInf::Ext) {
    SMLTokenType order[] = {SMLToken::DevInf::XNam,
                            SMLToken::DevInf::XVal};
    node->mustHaveChildren(1,order);
    node->orderChildren(2,order);
  } else
  if ((node->getType()) == SMLToken::DevInf::VerDTD) 
  {
  	//XXX TODO
	//if (version == ONE_ZERO)
		//node->setSTLValue("1.0");
	//else
	//	node->setSTLValue("1.1")
  } else
  if ((node->getType()) == SMLToken::DevInf::XNam) {
    mustBeTextNode(node,"");
  } else
  if ((node->getType()) == SMLToken::DevInf::XVal) {
    mustBeTextNode(node,"");
  }

  SMLTreeNode::nodeList *c = node->getChildren();
  for (SMLTreeNode::nodeList::iterator i=c->begin();i!=c->end();i++) 
    enforceConstraints((*i));
}

void SMLDevInf::save(QTextStream *o) {
  
  // An empty namespace container is sufficient, since there should be no
  // alien tokens withou our device specification!
  enforceConstraints();
  SMLNamespaceContainer *c = new SMLNamespaceContainer(syncmlVersion); 
  SMLFlattener *s = new NiceSyncMLFlattener(o,c);
  headNode->flatten(s);
  delete s;
  delete c;
}

SMLTreeNode* SMLDevInf::getDevInfTree() {
  enforceConstraints();
  return headNode->getChildren()->front();
}

void SMLDevInf::addDevInfToPacket(SMLTreeNode *t) 
{
	new SMLTreeNode(new SMLTreeNode(t,SMLToken::Meta),SMLToken::MetInf::Type,SMLMimeType::devinfxml);
	SMLTreeNode *i = new SMLTreeNode(t,SMLToken::Item);
	
	QString devinfStr;
	if (syncmlVersion == ONE_ZERO)
		devinfStr = "./devinf10";
	else
		devinfStr = "./devinf11";

	SMLURI *fakeBase = new SMLURI();
	SMLURI s(devinfStr, QString(""), fakeBase);
	s.addToNode(new SMLTreeNode(i,SMLToken::Source));
	SMLTreeNode *d = new SMLTreeNode(i,SMLToken::Data);
	// FIXME
	d->addChild(new SMLTreeNode(getDevInfTree()));
	
	delete fakeBase;
}

SMLDevInf::~SMLDevInf() {
  delete headNode;
}

#define getAndReturnValue(len,path) \
  SMLTreeNode* t = headNode->getChildOfType(len,path); \
  if (t==NULL) return ""; else { \
    t->simplifyMe(); \
    return t->getSTLValue(); \
  } 

#define setValue(len,path,value) \
  SMLTreeNode *t = headNode; \
  SMLTreeNode *n = t; \
  for (int i=0;i<len;i++) { \
    n = t->getChildOfType(path[i]); \
    if (n==NULL) n = new SMLTreeNode(t,path[i]); \
    t = n; \
  } \
  t->simplifyMe(); \
  t->setSTLValue(value)

string SMLDevInf::getDevId() {
  SMLTokenType p[] = {SMLToken::DevInf::DevInf,SMLToken::DevInf::DevId}; 
  getAndReturnValue(2, p);
}

void SMLDevInf::setDevId() {
  setDevId(SMLCreateUID());
}

void SMLDevInf::setDevId(string DevId) {
  SMLTokenType p[] = {SMLToken::DevInf::DevInf,SMLToken::DevInf::DevId}; 
  setValue(2, p, DevId);
}


string SMLDevInf::getDevTyp() {
  SMLTokenType p[] = {SMLToken::DevInf::DevInf,SMLToken::DevInf::DevTyp}; 
  getAndReturnValue(2, p);
}

void SMLDevInf::setDevTyp(string s) {
  SMLTokenType p[] = {SMLToken::DevInf::DevInf,SMLToken::DevInf::DevTyp}; 
  setValue(2, p, s);
}

void SMLDevInf::setDevTyp(DefTypEnum e) {
  switch (e) {
  case pager:
    setDevTyp("pager"); break;
  case handheld:
    setDevTyp("handheld"); break;
  case pda:
    setDevTyp("pda"); break;
  case phone:
    setDevTyp("phne"); break;
  case smartphone:
    setDevTyp("smartphone"); break;
  case server:
    setDevTyp("server"); break;
  default:
    setDevTyp("workstation"); break;
  }
}


string SMLDevInf::getFwV() {
  SMLTokenType p[] = {SMLToken::DevInf::DevInf,SMLToken::DevInf::FwV}; 
  getAndReturnValue(2, p);
}

void SMLDevInf::setFwV(string s) {
  SMLTokenType p[] = {SMLToken::DevInf::DevInf,SMLToken::DevInf::FwV}; 
  setValue(2, p, s);
}


string SMLDevInf::getHwV() {
  SMLTokenType p[] = {SMLToken::DevInf::DevInf,SMLToken::DevInf::HwV}; 
  getAndReturnValue(2, p);
}

void SMLDevInf::setHwV(string s) {
  SMLTokenType p[] = {SMLToken::DevInf::DevInf,SMLToken::DevInf::HwV}; 
  setValue(2, p, s);
}


string SMLDevInf::getMan() {
  SMLTokenType p[] = {SMLToken::DevInf::DevInf,SMLToken::DevInf::Man}; 
  getAndReturnValue(2, p);
}

void SMLDevInf::setMan(string s) {
  SMLTokenType p[] = {SMLToken::DevInf::DevInf,SMLToken::DevInf::Man}; 
  setValue(2, p, s);
}


string SMLDevInf::getMod() {
  SMLTokenType p[] = {SMLToken::DevInf::DevInf,SMLToken::DevInf::Mod}; 
  getAndReturnValue(2, p);
}

void SMLDevInf::setMod(string s) {
  SMLTokenType p[] = {SMLToken::DevInf::DevInf,SMLToken::DevInf::Mod}; 
  setValue(2, p, s);
}


string SMLDevInf::getOEM() {
  SMLTokenType p[] = {SMLToken::DevInf::DevInf,SMLToken::DevInf::OEM}; 
  getAndReturnValue(2, p);
}

void SMLDevInf::setOEM(string s) {
  SMLTokenType p[] = {SMLToken::DevInf::DevInf,SMLToken::DevInf::OEM}; 
  setValue(2, p, s);
}


string SMLDevInf::getSwV() {
  SMLTokenType p[] = {SMLToken::DevInf::DevInf,SMLToken::DevInf::SwV}; 
  getAndReturnValue(2, p);
}

void SMLDevInf::setSwV(string s) {
  SMLTokenType p[] = {SMLToken::DevInf::DevInf,SMLToken::DevInf::SwV}; 
  setValue(2, p, s);
}


string SMLDevInf::getVerDTD() {
  SMLTokenType p[] = {SMLToken::DevInf::DevInf,SMLToken::DevInf::VerDTD}; 
  getAndReturnValue(2, p);
}

void SMLDevInf::setVerDTD(SMLVersion version) 
{
	syncmlVersion = version;
	SMLTokenType p[] = {SMLToken::DevInf::DevInf,SMLToken::DevInf::VerDTD}; 
	if (version == ONE_ZERO)
		{ setValue(2, p, "1.0"); } //XXX FIXME this macro is very badly written
	else
		{ setValue(2, p, "1.1"); }
}

inline bool realHas(SMLTreeNode *headNode, SMLTokenType
                    nameSub, string Name) {
  SMLTreeNode *dNode = headNode->getChildOfType(SMLToken::DevInf::DevInf);
  for (SMLTreeNode::nodeList::iterator
         i=dNode->getChildren()->begin();i!=dNode->getChildren()->end();i++) {
    SMLTreeNode *eName = (*i)->getChildOfType(nameSub);
    if (eName!=NULL) {
      eName->simplifyMe();
      kdDebug() << " we have here : " << eName->getValue() << endl;
      if (eName->getSTLValue()==Name) return true;
    }
  }
  return false;
}

bool SMLDevInf::hasExt(string extName) {
  return realHas(headNode,SMLToken::DevInf::XNam,extName);
}

void SMLDevInf::setExtValues(string extName, vector<string> extValues) {
  deleteExt(extName);
  SMLTreeNode *dNode = headNode->getChildOfType(SMLToken::DevInf::DevInf);
  SMLTreeNode *eNode = new SMLTreeNode(dNode,SMLToken::DevInf::Ext);
  new SMLTreeNode(eNode,SMLToken::DevInf::XNam,QString(extName.c_str()));
  for (vector<string>::iterator i=extValues.begin();i!=extValues.end();i++) {
    new SMLTreeNode(eNode,SMLToken::DevInf::XVal,QString((*i).c_str()));
  }
}

vector<string> SMLDevInf::getExtValues(string extName) {
  vector<string> v;
  SMLTreeNode *dNode = headNode->getChildOfType(SMLToken::DevInf::DevInf);
  for (SMLTreeNode::nodeList::iterator
         i=dNode->getChildren()->begin();i!=dNode->getChildren()->end();i++) {
    SMLTreeNode *eName = (*i)->getChildOfType(SMLToken::DevInf::XNam);
    if (eName!=NULL) {
      eName->simplifyMe();
      if (eName->getSTLValue()==extName) {
        for (SMLTreeNode::nodeList::iterator
               j=(*i)->getChildren()->begin();j!=(*i)->getChildren()->end();j++)
          if ((*j)->getType()==SMLToken::DevInf::XVal) {
            (*j)->simplifyMe();
            v.push_back((*j)->getSTLValue());
          }
      }
    }
  }
  return v;
}

vector<string> realGet(SMLTreeNode *headNode, SMLTokenType nameSub) {
  SMLTreeNode *dNode = headNode->getChildOfType(SMLToken::DevInf::DevInf);
  vector<string> v;

  for (SMLTreeNode::nodeList::iterator
         i=dNode->getChildren()->begin();i!=dNode->getChildren()->end();i++) {
    SMLTreeNode *eName = (*i)->getChildOfType(nameSub);
    if (eName!=NULL) {
      eName->simplifyMe();
      v.push_back(eName->getSTLValue());
    }
  }
  return v;
}

vector<string> SMLDevInf::getExts() {
  return realGet(headNode,SMLToken::DevInf::XNam);
}

inline void realDelete(SMLTreeNode *headNode, SMLTokenType nameSub, string Name) {
  SMLTreeNode *dNode = headNode->getChildOfType(SMLToken::DevInf::DevInf);

  for (SMLTreeNode::nodeList::iterator
         i=dNode->getChildren()->begin();i!=dNode->getChildren()->end();) {

    SMLTreeNode *eName = (*i)->getChildOfType(nameSub);
    if (eName!=NULL) {
      eName->simplifyMe();
      if (eName->getSTLValue()==Name) {
        i=dNode->getChildren()->erase(i);
      } else i++;
    } else i++;
  }
}

void SMLDevInf::deleteExt(string extName) {
  realDelete(headNode,SMLToken::DevInf::XNam, extName);
}

bool SMLDevInf::hasDataStoreInfo(string SourceRef) {
  return realHas(headNode,SMLToken::DevInf::SourceRef,SourceRef);
}

void SMLDevInf::addDataStoreInfo(KSyncML::SyncMLDatabase::DataStoreType & values) {
  deleteDataStoreInfo(values.SourceRef);
  SMLTreeNode *dNode = headNode->getChildOfType(SMLToken::DevInf::DevInf);
  SMLTreeNode *eNode = new SMLTreeNode(dNode,SMLToken::DevInf::DataStore);

  new SMLTreeNode(eNode,SMLToken::DevInf::SourceRef, QString(values.SourceRef.c_str()));
  new SMLTreeNode(eNode,SMLToken::DevInf::DisplayName, QString(values.DisplayName.c_str()));

  if (values.MaxGUIDSize>0) new SMLTreeNode(eNode,SMLToken::DevInf::MaxGUIDSize,SMLUIntToStr(values.MaxGUIDSize));

  SMLTreeNode *rpNode = new SMLTreeNode(eNode,SMLToken::DevInf::RxPref);
  new SMLTreeNode(rpNode,SMLToken::DevInf::CTType,values.RxPref.first);
  new SMLTreeNode(rpNode,SMLToken::DevInf::VerCT,values.RxPref.second);

  SMLTreeNode *rNode = new SMLTreeNode(eNode,SMLToken::DevInf::Rx);
  for (multimap<string,string>::iterator
         i=values.Rx.begin();i!=values.Rx.end();i++) {
    new SMLTreeNode(rNode,SMLToken::DevInf::CTType,(*i).first);
    new SMLTreeNode(rNode,SMLToken::DevInf::VerCT,(*i).second);
  }

  SMLTreeNode *tpNode = new SMLTreeNode(eNode,SMLToken::DevInf::TxPref);
  new SMLTreeNode(tpNode,SMLToken::DevInf::CTType,values.TxPref.first);
  new SMLTreeNode(tpNode,SMLToken::DevInf::VerCT,values.TxPref.second);

  SMLTreeNode *tNode = new SMLTreeNode(eNode,SMLToken::DevInf::Tx);
  for (multimap<string,string>::iterator
         i=values.Tx.begin();i!=values.Tx.end();i++) {
    new SMLTreeNode(tNode,SMLToken::DevInf::CTType,(*i).first);
    new SMLTreeNode(tNode,SMLToken::DevInf::VerCT,(*i).second);
  }

  SMLTreeNode *mNode =  new SMLTreeNode(eNode,SMLToken::DevInf::DSMem);
  if (values.SharedMem) 
  	new SMLTreeNode(mNode,SMLToken::DevInf::SharedMem);
  if (values.MaxMem>0) 
  	new SMLTreeNode(mNode,SMLToken::DevInf::MaxMem,SMLUIntToStr(values.MaxMem));
  if (values.MaxID>0) 
  	new SMLTreeNode(mNode,SMLToken::DevInf::MaxID,SMLUIntToStr(values.MaxID));
  
  SMLTreeNode *sNode = new SMLTreeNode(eNode,SMLToken::DevInf::SyncCap);
  
  for (KSyncML::SyncMLDatabase::SyncCapType::iterator i=values.SyncCap.begin();i!=values.SyncCap.end();i++) {
      if ((*i).second) {
      new SMLTreeNode(sNode,SMLToken::DevInf::SyncType,(*i).first);
    }
    
  }
}

#define assignIfPresent(head,token,toAssign,convertFunc) { \
  SMLTreeNode *node = (head)->getChildOfType(token); \
  if (node!=NULL) { \
    node->simplifyMe(); \
    toAssign = convertFunc(node->getSTLValue()); \
  } \
}
 
KSyncML::SyncMLDatabase::DataStoreType SMLDevInf::getDataStoreInfo(string sourceRef) {
  KSyncML::SyncMLDatabase::DataStoreType store;
  SMLTreeNode *dNode = headNode->getChildOfType(SMLToken::DevInf::DevInf);
  for (SMLTreeNode::nodeList::iterator
         i=dNode->getChildren()->begin();i!=dNode->getChildren()->end();i++) {
    SMLTreeNode *eName = (*i)->getChildOfType(SMLToken::DevInf::SourceRef);
    if (eName!=NULL) {
      eName->simplifyMe();
      if (eName->getSTLValue()==sourceRef) {
        store.SourceRef = eName->getSTLValue();
        assignIfPresent((*i),SMLToken::DevInf::DisplayName,store.DisplayName,);
        assignIfPresent((*i),SMLToken::DevInf::MaxGUIDSize,store.MaxGUIDSize,SMLStrToUInt);

        SMLTreeNode *rpNode = (*i)->getChildOfType(SMLToken::DevInf::RxPref);
        if (rpNode!=NULL) {
          assignIfPresent(rpNode,SMLToken::DevInf::CTType,store.RxPref.first,);
          assignIfPresent(rpNode,SMLToken::DevInf::VerCT,store.RxPref.second,);
        }

        SMLTreeNode *sNode = (*i)->getChildOfType(SMLToken::DevInf::Rx);
        if (sNode!=NULL) {
          string CTType;
          for (SMLTreeNode::nodeList::iterator i=sNode->getChildren()->begin();
               i!=sNode->getChildren()->end(); i++) {
            (*i)->simplifyMe();
            if ((*i)->getType() == SMLToken::DevInf::CTType)
              CTType=(*i)->getSTLValue(); else
            if ((*i)->getType() == SMLToken::DevInf::VerCT)
              store.Rx.insert(pair<string,string>(CTType,(*i)->getSTLValue()));
          }
        }

        SMLTreeNode *tpNode = (*i)->getChildOfType(SMLToken::DevInf::TxPref);
        if (tpNode!=NULL) {
          assignIfPresent(tpNode,SMLToken::DevInf::CTType,store.TxPref.first,);
          assignIfPresent(tpNode,SMLToken::DevInf::VerCT,store.TxPref.second,);
        }

        SMLTreeNode *tNode = (*i)->getChildOfType(SMLToken::DevInf::Tx);
        if (tNode!=NULL) {
          string CTType;
          for (SMLTreeNode::nodeList::iterator i=tNode->getChildren()->begin();
               i!=tNode->getChildren()->end(); i++) {
            (*i)->simplifyMe();
            if ((*i)->getType() == SMLToken::DevInf::CTType)
              CTType=(*i)->getSTLValue(); else
            if ((*i)->getType() == SMLToken::DevInf::VerCT)
              store.Tx.insert(pair<string,string>(CTType,(*i)->getSTLValue()));
          }
        }

        SMLTreeNode *dNode = (*i)->getChildOfType(SMLToken::DevInf::DSMem);
        if (dNode != NULL) {
          store.SharedMem = dNode->hasChildOfType(SMLToken::DevInf::SharedMem);
          assignIfPresent(dNode,SMLToken::DevInf::MaxMem,store.MaxMem,SMLStrToUInt);
          assignIfPresent(dNode,SMLToken::DevInf::MaxMem,store.MaxID,SMLStrToUInt);
        }

        sNode = (*i)->getChildOfType(SMLToken::DevInf::SyncCap);
        if (sNode!=NULL) {
          for (SMLTreeNode::nodeList::iterator i=sNode->getChildren()->begin();
               i!=sNode->getChildren()->end(); i++) {
            (*i)->simplifyMe();
            store.SyncCap[(*i)->getSTLValue()] = true;
          }
        }
      }
    }
  }
  return store;
}

vector<string> SMLDevInf::getDataStores() {
  return realGet(headNode, SMLToken::DevInf::SourceRef);
}

void SMLDevInf::deleteDataStoreInfo(string DataStore) {
  realDelete(headNode, SMLToken::DevInf::SourceRef,DataStore);
}

SMLTokenType CTCapPath[] = {SMLToken::DevInf::DevInf,SMLToken::DevInf::CTCap}; 

#define getCTCapPath(ifnull) \
  SMLTreeNode *ctNode = headNode->getChildOfType(2,CTCapPath); \
  if (ctNode==NULL) ifnull;

bool SMLDevInf::hasCTCap(string CTType) {
  getCTCapPath(return false);
  for (SMLTreeNode::nodeList::iterator
         i=ctNode->getChildren()->begin();i!=ctNode->getChildren()->end();i++) {
    (*i)->simplifyMe();
    if (((*i)->getType()==SMLToken::DevInf::CTType) && ((*i)->getSTLValue()==CTType)) return true;
  }
  return false;
}

#define ValEnumOrDataType(node,param) { \
  if (!param.ValEnum.empty()) { \
    for (list<string>::iterator \
           x=param.ValEnum.begin();x!=param.ValEnum.end();x++) \
      new SMLTreeNode(node,SMLToken::DevInf::ValEnum,(*x)); \
  } else if (param.DataType!="") { \
    new SMLTreeNode(node,SMLToken::DevInf::DataType,param.DataType); \
    if (param.Size>0) \
      new SMLTreeNode(node,SMLToken::DevInf::Size,SMLUIntToStr(param.Size)); \
  } \
}

void SMLDevInf::setCTCap(CTCap values) {
  getCTCapPath(ctNode=new
               SMLTreeNode(headNode->getChildOfType(SMLToken::DevInf::DevInf),SMLToken::DevInf::CTCap));
  // deleteCTCap(values.CTType);
  new SMLTreeNode(ctNode,SMLToken::DevInf::CTType,values.CTType); 
 
  for (list<Prop>::iterator
         i=values.PropList.begin();i!=values.PropList.end();i++) {
    new SMLTreeNode(ctNode,SMLToken::DevInf::PropName,(*i).Name);
    ValEnumOrDataType(ctNode,(*i));
    new SMLTreeNode(ctNode,SMLToken::DevInf::DisplayName,(*i).DisplayName);
    for (list<Param>::iterator
           j=(*i).ParamList.begin();j!=(*i).ParamList.end();j++) {
      new SMLTreeNode(ctNode,SMLToken::DevInf::ParamName,(*j).Name);
      ValEnumOrDataType(ctNode,(*j));
      new SMLTreeNode(ctNode,SMLToken::DevInf::DisplayName,(*j).DisplayName);
    }
    
  }
  
}

inline bool getCTType(SMLTreeNode *ctNode,string CTType,
                 SMLTreeNode::nodeList::iterator *b,
                 SMLTreeNode::nodeList::iterator *e) {
  bool haveBegin=false;
  for (SMLTreeNode::nodeList::iterator
         i=ctNode->getChildren()->begin();i!=ctNode->getChildren()->end();i++) {
    (*i)->simplifyMe();
    if ((*i)->getType()==SMLToken::DevInf::CTType) {
      if (haveBegin) {
        *e = i;
        return true;
      }
      if ((*i)->getSTLValue()==CTType) {
        haveBegin=true;
        *b=i;
      }
    }
  } 
  if (!haveBegin) *b= ctNode->getChildren()->end();
  *e = ctNode->getChildren()->end();
  return haveBegin;
}

#define handleLastProp \
  if (lastProp != NULL) { \
    cap.PropList.push_back(*lastProp); \
    delete lastProp; \
  }
  
#define handleLastParam \
  if (lastParam != NULL) { \
    lastProp->ParamList.push_back(*lastParam); \
    delete lastParam; \
    lastParam=NULL; \
  }
  

SMLDevInf::CTCap SMLDevInf::getCTCap(string CTType) {
  CTCap cap;
  getCTCapPath(return cap);
  SMLTreeNode::nodeList::iterator b,e;
  Param *lastParam = NULL;
  Prop *lastProp = NULL;
  Param *last = NULL;
  if (getCTType(ctNode,CTType,&b,&e)) 
    for (SMLTreeNode::nodeList::iterator i=b;i!=e;i++) {
      if ((*i)->getType()==SMLToken::DevInf::CTType) {
        cap.CTType = (*i)->getSTLValue();
      } else if ((*i)->getType()==SMLToken::DevInf::PropName) {
        handleLastParam;
        handleLastProp;
        lastProp = new Prop();
        lastProp->Name = (*i)->getSTLValue();
        last = lastProp;
      } else if ((*i)->getType()==SMLToken::DevInf::ValEnum) {
        last->ValEnum.push_back((*i)->getSTLValue());
      } else if ((*i)->getType()==SMLToken::DevInf::DataType) {
        last->DataType=(*i)->getSTLValue();
      } else if ((*i)->getType()==SMLToken::DevInf::Size) {
        last->Size=SMLStrToUInt((*i)->getSTLValue());
      } else if ((*i)->getType()==SMLToken::DevInf::DisplayName) {
        last->DisplayName=(*i)->getSTLValue();
      } else if ((*i)->getType()==SMLToken::DevInf::ParamName) {
        handleLastParam;
        lastParam = new Param();
        lastParam->Name = (*i)->getSTLValue();
        last = lastParam;
      }
    } 
  handleLastParam;
  handleLastProp;
  return cap;
}

vector<string> SMLDevInf::getCTCaps() {
  vector<string> v;
  getCTCapPath(return v);

  for (SMLTreeNode::nodeList::iterator
         i=ctNode->getChildren()->begin();i!=ctNode->getChildren()->end();i++) {
    (*i)->simplifyMe();
    if ((*i)->getType()==SMLToken::DevInf::CTType) v.push_back((*i)->getSTLValue());
  }

  return v;
}


void SMLDevInf::deleteCTCap(string CTType) {
  getCTCapPath(return);
  SMLTreeNode::nodeList::iterator b,e;
  if (getCTType(ctNode,CTType,&b,&e)) ctNode->getChildren()->erase(b,e);
}

