/***************************************************************************
                                 libSyncML
           Implementation of the SyncML protocol (www.syncml.org)
         Copyright (C) 2001, 2002 by Max Berger <max.berger@xslt.de>
 **************************************************************************/

/***************************************************************************
 *                                                                         *
 *  This library is free software; you can redistribute it and/or          *
 *  modify it under the terms of the GNU Lesser General Public             *
 *  License as published by the Free Software Foundation; either           *
 *  version 2.1 of the License, or (at your option) any later              *
 *  version.                                                               *
 *                                                                         *
 *  This library is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU      *
 *  Lesser General Public License for more details.                        *
 *                                                                         *
 *  You should have received a copy of the GNU Lesser General Public       *
 *  License along with this library; if not, write to the Free             *
 *  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA      *
 *  02111-1307 USA                                                         *
 *                                                                         *
 ***************************************************************************/

#include "syncml.h"
#include "syncmlresponsepacket.h"
#include "xmlbeautifier.h" 
#include "xmlbinary.h"

#include <kdebug.h>


SMLResponsePacket::SMLResponsePacket() : headNode(NULL), header(NULL), body(NULL), nsCont(NULL), sourceDev(), targetDev()
{
	#ifdef DEBUG_MEM
	kdDebug() << "DEBUG_MEM - SMLResponsePacket: created." << endl;
	#endif
  
	messageID=1;
	CmdID=1;
}

SMLResponsePacket::~SMLResponsePacket() 
{
	#ifdef DEBUG_MEM
	kdDebug() << "DEBUG_MEM - SMLResponsePacket: destroyed." << endl;
	#endif

	delete nsCont;
	delete headNode;
}


void SMLResponsePacket::reinitialize() 
{
  nsCont = new SMLNamespaceContainer(syncmlVersion);
  
  headNode = new SMLTreeNode(NULL,SMLToken::Root);
  SMLTreeNode *root = new SMLTreeNode(headNode,SMLToken::SyncML);
  header = new SMLTreeNode(root,SMLToken::SyncHdr);
  
  if (syncmlVersion == ONE_ZERO)
  {
  	new SMLTreeNode(header,SMLToken::VerDTD,QString("1.0"));
  	new SMLTreeNode(header,SMLToken::VerProto,QString("SyncML/1.0"));
  }
  else
  {
  	new SMLTreeNode(header,SMLToken::VerDTD,QString("1.1"));
  	new SMLTreeNode(header,SMLToken::VerProto,QString("SyncML/1.1"));
  }
  // FIXME!
  new SMLTreeNode(header,SMLToken::SessionID,QString(SMLCreateUID().c_str()));
  new SMLTreeNode(header,SMLToken::MsgID,SMLUIntToStr(messageID));
  new SMLTreeNode(header,SMLToken::Target);
  new SMLTreeNode(header,SMLToken::Source);

  //XXX FIXME: TEST  
  //new SMLTreeNode(header,SMLToken::RespURI, QString("http://169.254.1.68:8080?sessionid=8297a8161437766c"));
  
  body = new SMLTreeNode(root,SMLToken::SyncBody);
  sendAtAll=false;
}


void SMLResponsePacket::setSyncMLVersion(SMLVersion protocolVersion)
{
	syncmlVersion  = protocolVersion;
	reinitialize();
}


bool SMLResponsePacket::send(SMLFlattener *flat) {

  if (!sendAtAll) return false;

  // Now add all the extra stuff that could have been changed.
  // This is done now that that we don't need the orderchildren function

  kdDebug() << "SMLResponsePacket: sending. So far we are: " << endl;
  

  if (!header)  
  {
  	kdWarning() << "SMLResponsePacket: no header!" << endl;
	return false;
  }  
  
  //kdDebug() << "SyncMLSession: ======== DUMP HEADER ========" << endl;
  //header->dump();
  //kdDebug() << "SyncMLSession: ======== DUMP HEADER END ====" << endl;
  
  

  
  SMLTreeNode * sourceNode = header->getChildOfType(SMLToken::Source);
  SMLTreeNode * targetNode = header->getChildOfType(SMLToken::Target);
  
  if (sourceNode)  
  	sourceDev.addToNode(sourceNode);
  else
  	kdWarning() << "SMLResponsePacket: no source node" << endl;

  if (targetNode)
	targetDev.addToNode(targetNode);
  else
  	kdWarning() << "SMLResponsePacket: no target node" << endl;
	
  kdDebug() << "SMLResponsePacket: elimitate empty children." << endl;
  header->eliminateEmptyChildren();

  new SMLTreeNode(body,SMLToken::Final);

  kdDebug() << "SMLResponsePacket: flatten." << endl;
  headNode->flatten(flat);
  kdDebug() << "SMLResponsePacket: flush." << endl;
  flat->flush();
  
  string sessionID=getSessionID();
  kdDebug() << "SMLResponsePacket: reinitialize. sessionID=" << sessionID.c_str() << endl;
  
  delete headNode;
  messageID++;
  reinitialize();
  setSessionID(sessionID);
  return true;
}

SMLTreeNode* SMLResponsePacket::addSubCommand(SMLTreeNode* where, SMLTokenType typ, SMLCmdRef* cmdid, bool vitallyForSend) 
{
  sendAtAll|=vitallyForSend;
  SMLTreeNode *t = new SMLTreeNode(where,typ);
  new SMLTreeNode(t,SMLToken::CmdID,SMLUIntToStr(CmdID));
  cmdid->CmdRef = CmdID;
  cmdid->MsgRef = messageID;
  CmdID++;
  return t;
}

SMLTreeNode* SMLResponsePacket::addCommand(SMLTokenType typ, SMLCmdRef* cmdid, bool vitallyForSend) 
{
	
	//if (vitallyForSend) 
	//	kdDebug() << "SMLResponsePacket: adding a vital command." << endl;
	//else
	//	kdDebug() << "SMLResponsePacket: adding a non-vital command." << endl;
	return addSubCommand(body,typ,cmdid,vitallyForSend);
}

string SMLResponsePacket::getSessionID() {
  SMLTreeNode* t = header->getChildOfType(SMLToken::SessionID);
  return t->getSTLValue();
}

void SMLResponsePacket::setSessionID(string sessionID) {
  SMLTreeNode* t = header->getChildOfType(SMLToken::SessionID);
  t->setSTLValue(sessionID);
}

bool nodeIsHeader(SMLTreeNode *node) {
  return (node->getType()==SMLToken::SyncHdr);
}

SMLTreeNode* addMsgRefAndCmdRef(SMLTreeNode *node, SMLUInt MsgRef, SMLTreeNode *nodeRef) {
	new SMLTreeNode(node,SMLToken::MsgRef,SMLUIntToStr(MsgRef));
	string CmdRef;
	SMLTreeNode *n;
	SMLTreeNode *refCmdNode = nodeRef;
	if (nodeIsHeader(nodeRef)) CmdRef="0"; else
		if ( (n=nodeRef->getChildOfType(SMLToken::CmdID))!=NULL) 
			CmdRef=n->getSimplifiedSTLValue(); 
		else
			if ( (n=nodeRef->getParent()->getChildOfType(SMLToken::CmdID))!=NULL) 
			{
				CmdRef=n->getSimplifiedSTLValue();
				refCmdNode=nodeRef->getParent();
			}
	new SMLTreeNode(node,SMLToken::CmdRef,QString(CmdRef.c_str()));
	return refCmdNode;
}

SMLTreeNode* SMLResponsePacket::addStatus(SMLUInt MsgRef, SMLTreeNode *nodeRef, SMLErrorType statusCode) 
{
 // kdDebug() << " SMLResponsePacket: adding a new status" << endl;
  SMLCmdRef ret;
  SMLTreeNode *s = addCommand(SMLToken::Status,&ret,!nodeIsHeader(nodeRef));
  SMLTreeNode *cmr = addMsgRefAndCmdRef(s,MsgRef,nodeRef);
  new SMLTreeNode(s,SMLToken::Cmd, QString(WBXTranslator::getSyncMLStringForToken(cmr->getType()).c_str()));
  // FIXME
  new SMLTreeNode(s,SMLToken::Data,SMLUIntToStr(statusCode));
  return s;
}

SMLTreeNode* SMLResponsePacket::addStatus(SMLUInt MsgRef, SMLUInt CmdRef, QString cmd, SMLErrorType statusCode)
{
	//kdDebug() << " SMLResponsePacket: adding a status reply." << endl;
	SMLCmdRef ret;
	SMLTreeNode *s = addCommand(SMLToken::Status, &ret, true);
	if (MsgRef != 0)
		new SMLTreeNode(s, SMLToken::MsgRef, SMLUIntToStr(MsgRef));
	
	if (CmdRef != 0)
		new SMLTreeNode(s, SMLToken::CmdRef, SMLUIntToStr(CmdRef));
	
	new SMLTreeNode(s,SMLToken::Cmd, cmd);
	
	new SMLTreeNode(s,SMLToken::Data,SMLUIntToStr(statusCode));
	return s;

}


SMLTreeNode* SMLResponsePacket::addResults(SMLUInt MsgRef, SMLTreeNode *nodeRef) 
{
	//kdDebug() << "SMLResponsePacket: adding result for msg " << MsgRef << endl;
  	SMLCmdRef ret;
  	SMLTreeNode *s = addCommand(SMLToken::Results,&ret,!nodeIsHeader(nodeRef));
  	addMsgRefAndCmdRef(s,MsgRef,nodeRef);
  	return s;
}
