/***************************************************************************
                                 libSyncML
           Implementation of the SyncML protocol (www.syncml.org)
         Copyright (C) 2001, 2002 by Max Berger <max.berger@xslt.de>
 **************************************************************************/

/***************************************************************************
 *                                                                         *
 *  This library is free software; you can redistribute it and/or          *
 *  modify it under the terms of the GNU Lesser General Public             *
 *  License as published by the Free Software Foundation; either           *
 *  version 2.1 of the License, or (at your option) any later              *
 *  version.                                                               *
 *                                                                         *
 *  This library is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU      *
 *  Lesser General Public License for more details.                        *
 *                                                                         *
 *  You should have received a copy of the GNU Lesser General Public       *
 *  License along with this library; if not, write to the Free             *
 *  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA      *
 *  02111-1307 USA                                                         *
 *                                                                         *
 ***************************************************************************/

#include "xmlparser_tree.h"
#include "xmlbeautifier.h"
#include <kdebug.h>


SMLTreeNode::SMLTreeNode(SMLTreeNode* parentNode, SMLTokenType type, std::string val)
{
  myType = type;
  parent = parentNode; // Happens again in addChild, is only here for parent==NULL;
  value = QString(val.c_str());
  if (parentNode != NULL) {
    parentNode->addChild(this);
  }

}


SMLTreeNode::SMLTreeNode(SMLTreeNode* parentNode, SMLTokenType type, QString val) {
  myType = type;
  parent = parentNode; // Happens again in addChild, is only here for parent==NULL;
  value = val;
  if (parentNode != NULL) {
    parentNode->addChild(this);
  }
}

SMLTreeNode::SMLTreeNode(SMLTreeNode *copyMe) {
  parent=NULL;
  myType=copyMe->myType;
  value=copyMe->value;

  for (nodeList::iterator i=copyMe->children.begin();i!=copyMe->children.end();i++) {
    //SMLTreeNode *newChild = 
	new SMLTreeNode(*i);
    this->addChild(*i);
  }
}

SMLTreeNode* SMLTreeNode::getParent() { 
  return parent; 
}

string SMLTreeNode::getSTLValue() {
  return string(value.ascii());
}

string SMLTreeNode::getSimplifiedSTLValue() {
  simplifyMe();
  return getSTLValue();
}

void SMLTreeNode::setSTLValue(string v) {
  value = QString(v.c_str());
}

SMLTreeNode::nodeList* SMLTreeNode::getChildren() {
  return &children;
}

SMLTokenType SMLTreeNode::getType () { 
  return myType; 
}

void SMLTreeNode::addChild(SMLTreeNode *newChild) {
  newChild->parent=this;
  children.push_back(newChild);
}

bool SMLTreeNode::simplifyMe() 
{
  if (children.empty()) return false;
  // FIXME: This can be done more efficient by checking whether the first
  // child has a sibling;
  if (children.size()!=1) return false; 
  
  SMLTreeNode *n = children.front();
  
  //if (n)
  //	kdDebug() << "SMLTreeNode::simplifyMe: n is " << n->getValue() << endl;
  
  if ((n->getType()) != (const uint16_t)SMLToken::String) return false;
  if ((myType.space != SMLNamespace::SyncML) &&
      (myType.space != SMLNamespace::MetInf) &&
      (myType.space != SMLNamespace::DevInf))
    return false;
  value = n->value;
  delete n;
  children.clear();
  return true;
}

void SMLTreeNode::dump(int indentation)
{
	//horribly ineffective...
	QString ind;
	for (int i = 0; i < indentation; i++)
		ind.append(" ");
		
	if (children.empty())
	{
		kdDebug() << ind << "No children. My value: " << getValue() << endl;
	}
	else
	{
		kdDebug() << ind << "My children start" << endl;
		for (nodeList::iterator i = children.begin(); i != children.end(); i++) 
		{
			(*i)->dump(indentation + 1);
		}
		kdDebug() << ind << "My children end" << endl;
	}
	
}

bool SMLTreeNode::hasChildOfType(SMLTokenType type) {
  for (nodeList::iterator i=children.begin();i!=children.end();i++) {
    if ((*i)->getType() == type) return true;
  }
  return false;
}

SMLTreeNode* SMLTreeNode::getChildOfType(SMLUInt ocount, SMLTokenType *path) {
  if (ocount==0) return this;
  //if (children.empty())
  //	kdWarning() << "SMLTreeNode children empty" << endl;
  for (nodeList::iterator i=children.begin();i!=children.end();i++) {
    if ((*i)->getType() == path[0]) return (*i)->getChildOfType(ocount-1,&path[1]);
  }
  return NULL;
}

bool SMLTreeNode::flatten(SMLFlattener *flat) {
  
  // Is the operation successfull ?
  bool result = true;

  // Do we need a start and End tag ? (not needed for special nodes)
  bool startAndEnd = true;

  // We don't output illegal tokens
  if (myType == SMLToken::Illegal) return false;

  if (myType == SMLToken::Root) {
    // Root Token: output xml processing instructions such as <?xml> and stuff.
    result&=flat->header();
    startAndEnd = false;
  } else if (myType == SMLToken::String) {
    // A plain string node. 
    startAndEnd = false;
  }

  SMLTokenSpaceType parentSpace = SMLNamespace::None;
  if (parent != NULL) parentSpace = parent->getType().space;

  string stlValue = getSTLValue();
  // empty node?
  if ((startAndEnd)&&(isEmpty())) {
    result&=flat->emptyTag(myType,parentSpace,&stlValue);
    startAndEnd=false;
  }

  // Output Start Tag if needed
  if (startAndEnd) {
    result&=flat->startTag(myType,parentSpace,&stlValue);
  }

  // If this is a string or a simplified node, we have a string and no
  // children
  if ((getSTLValue().length() > 0)&&(myType.token!=SMLToken::AlienToken)) {
    result&=flat->text(&stlValue);
  } else {
    // Else do all children now
    for (nodeList::iterator i=children.begin();i!=children.end();i++)
      (*i)->flatten(flat);
  }

  // and the end-tag
  if (startAndEnd) result&=flat->endTag(myType,&stlValue);

  // and return wheter we were sucessfull
  return result;
}

void SMLTreeNode::orderChildren(SMLUInt ocount,SMLTokenType *order) {

  nodeList* bucket = new nodeList[ocount];

  map<SMLTokenType,SMLUInt> BucketForToken;
  for (SMLUInt i=0;i<ocount;i++) {
    BucketForToken[order[i]]=i;
  }

  for (nodeList::iterator i=children.begin();i!=children.end();i++) {
    bucket[BucketForToken[(*i)->myType]].push_back((*i));
    (*i)=NULL;
  }

  children.clear();
  
  for (SMLUInt i=0;i<ocount;i++) {
    children.splice(children.end(),bucket[i]);
  }
  
  delete[] bucket;
}

void SMLTreeNode::mustHaveChildren(SMLUInt ocount,SMLTokenType *toHave) {
  map<SMLTokenType,bool> hasChild;
  for (SMLUInt i=0;i<ocount;i++) hasChild[toHave[i]]=false;
  for (nodeList::iterator i=children.begin();i!=children.end();i++)
    hasChild[(*i)->myType] = true;

  for (map<SMLTokenType,bool>::iterator i=hasChild.begin();i!=hasChild.end();i++) {
    if ((*i).second==false) {
      new SMLTreeNode(this,(*i).first);
    }
  }
}

#define reallyEliminate(condition) \
  for (nodeList::iterator i=children.begin();i!=children.end();) { \
    if (condition) \
      i = children.erase(i); \
    else \
      i++; \
    }
  
void SMLTreeNode::eliminateTextChildren() {
  reallyEliminate((*i)->myType==SMLToken::String);
}

bool SMLTreeNode::isEmpty() {
  return ((getSTLValue()=="")&&(children.empty()));
}

void SMLTreeNode::eliminateEmptyChildren() {
  // FIXME: Is there a better way to declare the valid "empty" tags in an
  // independent way?
  reallyEliminate( ((*i)->isEmpty()) &&
                   ((*i)->myType!=SMLToken::Final) &&
                   ((*i)->myType!=SMLToken::DevInf::SharedMem)
                   );
}

void SMLTreeNode::eliminateChildren() {
  for (nodeList::iterator
         i=children.begin();i!=children.end();i=children.erase(i));
}

SMLTreeNode::~SMLTreeNode() {
  eliminateChildren();
}
